def fit_into_bounds(
        image_width,
        image_height,
        frame_width,
        frame_height,
        only_shrink=False,
    ):
    '''
    Given the w+h of the image and the w+h of the frame,
    return new w+h that fits the image into the frame
    while maintaining the aspect ratio.

    (1920, 1080, 400, 400) -> (400, 225)
    '''
    width_ratio = frame_width / image_width
    height_ratio = frame_height / image_height
    ratio = min(width_ratio, height_ratio)

    new_width = int(image_width * ratio)
    new_height = int(image_height * ratio)

    if only_shrink and (new_width > image_width or new_height > image_height):
        return (image_width, image_height)

    return (new_width, new_height)

def rotate_by_exif(image):
    '''
    Rotate the image according to its exif data, so that it will display
    correctly even if saved without the exif.
    '''
    # Thank you Scabbiaza
    # https://stackoverflow.com/a/26928142
    import PIL.ExifTags

    for (ORIENTATION_KEY, val) in PIL.ExifTags.TAGS.items():
        if val == 'Orientation':
            break

    try:
        exif = image._getexif()
    except AttributeError:
        return image

    if exif is None:
        return image

    try:
        rotation = exif[ORIENTATION_KEY]
    except KeyError:
        return image

    if rotation == 3:
        image = image.rotate(180, expand=True)
    elif rotation == 6:
        image = image.rotate(270, expand=True)
    elif rotation == 8:
        image = image.rotate(90, expand=True)

    return image



# import pyperclip moved to stay lazy.
import os
import sys

CLIPBOARD_STRINGS = ['!c', '!clip', '!clipboard']
INPUT_STRINGS = ['!i', '!in', '!input', '!stdin']
EOF = '\x1a'

# In pythonw, stdin and stdout are None.
IN_PIPE = (sys.stdin is not None) and (not sys.stdin.isatty())
OUT_PIPE = (sys.stdout is not None) and (not sys.stdout.isatty())

class PipeableException(Exception):
    pass

class NoArguments(PipeableException):
    pass

def ctrlc_return1(function):
    '''
    Apply this decorator to your argparse gateways, and if the user presses
    ctrl+c then the gateway will return 1 as its status code without the
    stacktrace appearing.

    This helps me avoid wrapping the entire function in a try-except block.

    Don't use this if you need to perform some other kind of cleanup on ctrl+c.
    '''
    def wrapped(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except KeyboardInterrupt:
            return 1
    return wrapped

def _multi_line_input(prompt=None):
    if prompt is not None and not IN_PIPE:
        sys.stderr.write(prompt)
        sys.stderr.flush()

    while True:
        line = sys.stdin.readline()
        parts = line.split(EOF)
        line = parts[0]
        has_eof = len(parts) > 1

        # Note that just by hitting enter you always get \n, so this does NOT
        # mean that input finishes by submitting a blank line! It means that you
        # submitted EOF as the first character of a line, so there was nothing
        # in parts[0]. If EOF is in the middle of the line we'll still yield the
        # first bit before breaking the loop.
        if line == '':
            break

        line = line.rstrip('\n')
        yield line

        if has_eof:
            break

def multi_line_input(prompt=None):
    '''
    Yield multiple lines of input from the user, until they submit EOF.
    EOF is usually Ctrl+D on linux and Ctrl+Z on windows.

    The prompt is only shown for non-pipe situations, so you do not need to
    adjust your `prompt` argument for pipe/non-pipe usage.
    '''
    lines = _multi_line_input(prompt=prompt)
    if not IN_PIPE:
        # Wait until the user finishes all their lines before continuing.
        # The caller might be processing + printing these lines in a loop
        # and it would be weird if they start outputting before the user has
        # finished inputting.
        lines = list(lines)
    return lines

def input(
        arg,
        *,
        input_prompt=None,
        read_files=False,
        skip_blank=False,
        strip=False,
    ):
    '''
    Given an argument (probably from the command line), resolve it into a
    generator of lines.

    If the arg is in CLIPBOARD_STRINGS, the contents of the clipboard are taken
    and split into lines.
    If the arg is in INPUT_STRINGS, input is read from stdin with an optional
    input_prompt that is only shown during non-pipe input (actual typing).
    If read_files is True and the arg is the path to an existing file, the file
    is read as utf-8 lines.
    If none of the above, then the argument is taken literally.

    Resolution is not recursive: if the clipboard contains the name of a file,
    it won't be read, etc.

    In addition to the above resolution techniques, you also have the option to
    strip lines before yielding them, and skip lines which are emptystring (if
    strip is False, then all-whitespace lines will still be yielded). If you're
    modifying input but overall maintaining its original structure, you probably
    want these both False. If you're just crunching numbers you probably want
    them both True.

    So, your calling code should not have to make any adjustments -- just call
    this function however is appropriate for your data sink and enjoy.
    '''
    arg_lower = arg.lower()

    if arg_lower in INPUT_STRINGS:
        lines = multi_line_input(prompt=input_prompt)

    elif arg_lower in CLIPBOARD_STRINGS:
        import pyperclip
        lines = pyperclip.paste().splitlines()

    elif read_files and os.path.isfile(arg):
        lines = open(arg, 'r', encoding='utf-8')
        lines = (line.rstrip('\r\n') for line in lines)

    else:
        lines = arg.splitlines()

    for line in lines:
        if strip:
            line = line.strip()
        if skip_blank and not line:
            continue
        yield line

def _output(stream, line, end):
    line = str(line)
    stream.write(line)
    if not line.endswith(end):
        stream.write(end)
    if not OUT_PIPE:
        stream.flush()

def stdout(line, end='\n'):
    _output(sys.stdout, line, end)

def stderr(line, end='\n'):
    _output(sys.stderr, line, end)

# backwards compat
output = stdout

def go(args=None, *input_args, **input_kwargs):
    '''
    Automatically resolve all commandline arguments, or read from stdin if
    there are no arguments.

    This function is only useful if you have *no other arguments* besides your
    program's main data sink. You will not be able to use argparse in
    conjunction with this function. If you want to support more arguments,
    it's better to use a regular argparse argument which is then passed into
    pipeable.input to resolve it for your data sink.
    '''
    if args is None:
        args = sys.argv[1:]

    if not args:
        # There are no arguments, and...
        if IN_PIPE:
            # we are being piped to, so read the pipe.
            args = [INPUT_STRINGS[0]]
        else:
            # we are on the terminal, so cry for help.
            raise NoArguments()

    for arg in args:
        yield from input(arg, *input_args, **input_kwargs)



'''
On Windows, square brackets do not have a special meaning in glob strings.
However, python's glob module is written for unix-style globs in which brackets
represent character classes / ranges.

On Windows we should escape those brackets to get the right results.
But calling glob.escape would also escape asterisk which may not be desired.
So this module just provides a modified version of glob.glob which will escape
only square brackets when called on windows, and behave normally on linux.
'''
import fnmatch as python_fnmatch
import glob as python_glob
import os
import re

def fix(pattern):
    if os.name == 'nt':
        pattern = re.sub(r'(\[|\])', r'[\1]', pattern)
    return pattern

def fnmatch(name, pat):
    return python_fnmatch.fnmatch(name, fix(pat))

def glob(pathname, *, recursive=False):
    return python_glob.glob(fix(pathname), recursive=recursive)

def is_glob(pattern):
    '''
    Improvements can be made to consider [] ranges for unix, but properly
    parsing the range syntax is not something I'm interested in doing right now
    and it would become the largest function in the whole module.
    '''
    return any(c in pattern for c in '*?')



####################################################################################################
## ABC_CHOOSER #####################################################################################
####################################################################################################

def _abc_make_option_letters(options):
    import math
    import string
    from voussoirkit import basenumber
    option_letters = {}
    letter_length = math.ceil(len(options) / 26)
    for (index, option) in enumerate(options):
        letter = basenumber.to_base(index, 26, alphabet=string.ascii_lowercase).rjust(letter_length, 'a')
        option_letters[letter] = option

    return option_letters

def abc_chooser(options, prompt='', must_pick=False):
    '''
    Given a list of options, the user will pick one by the corresponding letter.
    The return value is the item from the options list, or None if must_pick is
    False and the user entered nothing.
    '''
    option_letters = _abc_make_option_letters(options)

    while True:
        message = []
        for (letter, option) in option_letters.items():
            message.append(f'{letter}. {option}')
        print('\n'.join(message))
        choice = input(prompt).strip().lower()

        if not choice:
            if must_pick:
                print()
                continue
            else:
                return

        if choice not in option_letters:
            print()
            continue

        return option_letters[choice]

def abc_chooser_many(options, prompt='', label='X'):
    '''
    Given a list of options, the user may pick zero or many options by their
    corresponding letter. They can toggle the options on and off as long as
    they like, and submit their selection by entering one more blank line.
    The return value is a list of items from the options list.
    '''
    selected = set()
    option_letters = _abc_make_option_letters(options)

    while True:
        message = []
        for (letter, option) in option_letters.items():
            this_label = label if letter in selected else ''
            this_label = this_label.center(len(label))
            message.append(f'{letter}. [{this_label}] {option}')
        print('\n'.join(message))

        choice = input(prompt).strip().lower()

        if not choice:
            break

        if choice not in option_letters:
            print()
            continue

        if choice in selected:
            selected.remove(choice)
        else:
            selected.add(choice)
        print()

    choices = [option_letters[letter] for letter in option_letters if letter in selected]
    return choices

####################################################################################################
## GETPERMISSION ###################################################################################
####################################################################################################

YES_STRINGS = ['yes', 'y']
NO_STRINGS = ['no', 'n']

def getpermission(
        prompt=None,
        *,
        yes_strings=YES_STRINGS,
        no_strings=NO_STRINGS,
        must_pick=False,
    ):
    '''
    Prompt the user with a yes or no question.

    Return True for yes, False for no, and None if undecided.

    You can customize the strings that mean yes or no. For example, you could
    create a "type the name of the thing you're about to delete" prompt.

    If `must_pick`, then undecided is not allowed and the input will repeat
    until they choose an acceptable answer. Either way, the intended usage of
    `if getpermission():` will always only accept in case of explicit yes.
    '''
    if prompt is not None:
        print(prompt)
    while True:
        answer = input(f'{yes_strings[0]}/{no_strings[0]}> ').strip()
        yes = answer.lower() in (option.lower() for option in yes_strings)
        no = answer.lower() in (option.lower() for option in no_strings)
        if yes or no or not must_pick:
            break

    if yes:
        return True
    if no:
        return False
    return None



import itertools

def chunk_generator(sequence, chunk_length):
    '''
    Given any sequence input, yield lists of length <= `chunk_length`.

    Note: this generator always yields lists, even if the input was a string.
    I don't want to deal with special cases of types that return differently.
    '''
    iterator = iter(sequence)
    while True:
        chunk = list(itertools.islice(iterator, chunk_length))
        if not chunk:
            break
        yield chunk



import os

def quote(arg):
    if os.name == 'nt':
        # If the command contains comma, semicolon, or equals, only the left
        # half is considered the command and the rest is considered the first
        # argument. If these characters are in other arguments they are
        # sometimes parsed as separate by builtin commands like del, but not
        # separate for external applications.
        # Ampersand, pipe, and caret are process flow and special escape.
        # Quotes inside quotes must be doubled up.
        badchars = [',', ';', '=', '&', '|', '^', '"']
        if arg == '' or any(c.isspace() for c in arg) or any(c in arg for c in badchars):
            arg = arg.replace('"', '""')
            arg = f'"{arg}"'
        return arg
    else:
        # Semicolon is command delimiter.
        # Equals assigns shell variables.
        # Quotes inside quotes must be escaped with backslash.
        badchars = [' ', ';', '=', '&', '|']
        if arg == '' or any(c.isspace() for c in arg) or any(c in arg for c in badchars):
            arg = arg.replace("'", "\\'")
            arg = f"'{arg}'"
        return arg

def format_command(command):
    cmd = [quote(x) for x in command]
    cmd = ' '.join(cmd)
    cmd = cmd.strip()
    return cmd

def print_command(command, prefix=''):
    print(prefix + format_command(command))



import re
import unicodedata

def collapse_whitespace(text):
    '''
    Replace all whitespace sequences with a single space and strip the ends.
    '''
    text = re.sub(r'\s+', ' ', text.strip())
    return text

def comma_space_split(text):
    '''
    Split the string by commas and spaces, discarding all extra
    whitespace and blank parts.

    'a b, c,,d' -> ['a', 'b', 'c', 'd']
    '''
    if text is None:
        return text
    return re.split(r'[ ,]+', text.strip())

def excise(text, mark_left, mark_right):
    '''
    Remove the text between the left and right landmarks, including the
    landmarks themselves, and return the rest of the text.

    excise('What a wonderful day [soundtrack].mp3', ' [', ']') ->
    returns 'What a wonderful day.mp3'
    '''
    if mark_left in text and mark_right in text:
        return text.split(mark_left, 1)[0] + text.rsplit(mark_right, 1)[-1]
    return text

def pascal_to_loudsnakes(text):
    '''
    PascalCase -> PASCAL_CASE
    HTMLDocument -> HTML_DOCUMENT
    '''
    text = re.sub(r'([a-z])([A-Z])', r'\1_\2', text)
    text = re.sub(r'([A-Z]+)([A-Z][a-z])', r'\1_\2', text)
    text = text.upper()
    return text

def remove_characters(text, characters):
    translator = {ord(c): None for c in characters}
    text = text.translate(translator)
    return text

def remove_control_characters(text):
    '''
    Thanks Alex Quinn
    https://stackoverflow.com/a/19016117

    unicodedata.category(character) returns some two-character string
    where if [0] is a C then the character is a control character.
    '''
    return ''.join(c for c in text if unicodedata.category(c)[0] != 'C')

def title_capitalize(text):
    text = text.strip().title()
    articles = [
        'a',
        'an',
        'and',
        'at',
        'for',
        'from',
        'in',
        'of',
        'on',
        'the',
        'to',
    ]
    for article in articles:
        text = re.sub(rf' {article}\b', f' {article.lower()}', text, flags=re.IGNORECASE)

    text = text.replace('\'S', '\'s')

    # Roman numerals. Not handling L, M yet because I don't want to mess up
    # real words like "mix", but let's take a look at expanding this in
    # the future.
    text = re.sub(r'(\b[ivx]+\b)', lambda m: m.group(1).upper(), text, flags=re.IGNORECASE)
    return text



import collections
import time

from voussoirkit import sentinel

NO_ITEM = sentinel.Sentinel('no item')

class Cache:
    def __init__(self, maxlen, expiry=float('inf')):
        self.maxlen = maxlen
        self.expiry = expiry
        self.cache = collections.OrderedDict()

        # To prevent excessive purge loops during repeated setitem, only allow
        # a purge once every this many seconds.
        self.max_purge_frequency = 0.5

        self._last_purge = 0

    def __contains__(self, key):
        return self.get(key, fallback=NO_ITEM) is not NO_ITEM

    def __getitem__(self, key):
        '''
        Return the key's value, or raise KeyError.
        '''
        # Let KeyError raise to caller.
        (value, timestamp) = self.cache.pop(key)

        now = time.time()
        if (now - timestamp) > self.expiry:
            raise KeyError(key)

        self.cache[key] = (value, timestamp)
        return value

    def __len__(self):
        '''
        Purge expired items, then count the length.
        Due to the purge, this method is not O(1) as most len methods are.
        '''
        self._purge_expired()
        return len(self.cache)

    def __setitem__(self, key, value):
        # If the key was already present, we don't need to worry about maxlen
        # because the net change is zero. If it was not present (KeyError) we
        # check the maxlen and pop the oldest item if needed.
        # Either way we update the timestamp.
        try:
            self.cache.pop(key)
        except KeyError:
            if len(self) >= self.maxlen:
                self.cache.popitem(last=False)
        self.cache[key] = (value, time.time())

    def _purge_expired(self):
        now = time.time()
        if now - self._last_purge < self.max_purge_frequency:
            return

        for (key, (value, timestamp)) in list(self.cache.items()):
            if (now - timestamp) > self.expiry:
                self.cache.pop(key)

        self._last_purge = now

    def clear(self):
        '''
        Remove everything from the cache.
        '''
        self.cache.clear()

    def get(self, key, fallback=None):
        '''
        Return the key's value, or fallback in case of KeyError.
        '''
        try:
            return self[key]
        except KeyError:
            return fallback

    def pop(self, key):
        '''
        Remove the key and return its value, or raise KeyError.
        '''
        (value, timestamp) = self.cache.pop(key)
        return value

    def remove(self, key):
        '''
        Remove the item and ignore KeyError.
        '''
        try:
            self.pop(key)
        except KeyError:
            pass



'''
This module forwards everything from logging, with the addition of a level
LOUD=1 and all loggers are given the `loud` method.
'''
from logging import *

_getLogger = getLogger

LOUD = 1

def getLogger(*args, **kwargs):
    log = _getLogger(*args, **kwargs)
    add_loud(log)
    return log

def add_loud(log):
    def loud(self, message, *args, **kwargs):
        if self.isEnabledFor(LOUD):
            self._log(LOUD, message, args, **kwargs)

    addLevelName(LOUD, 'LOUD')
    log.loud = loud.__get__(log, log.__class__)

def get_level_by_argv(argv):
    '''
    If any of the following arguments are present in argv, return the
    corresponding log level along with a new copy of argv that has had the
    argument string removed.

    Since we are removing the argument, your argparser should not have options
    with these same names.

    --loud: LOUD
    --debug: DEBUG
    --quiet: ERROR
    --silent: 99999999999
    none of the above: INFO
    '''
    argv = argv[:]

    if '--loud' in argv:
        level = LOUD
        argv.remove('--loud')
    elif '--debug' in argv:
        level = DEBUG
        argv.remove('--debug')
    elif '--quiet' in argv:
        level = ERROR
        argv.remove('--quiet')
    elif '--silent' in argv:
        level = 99999999999
        argv.remove('--silent')
    else:
        level = INFO

    return (level, argv)

def set_level_by_argv(log, argv):
    '''
    This function is helpful for single-file scripts where you instantiate the
    logger in the global scope, and use this function to set its level
    according to the "--debug" flags in argv, then pass the rest of argv to
    your argparser.
    '''
    basicConfig()

    (level, argv) = get_level_by_argv(argv)
    log.setLevel(level)

    return argv
